//	PicMonger automates retrieval, decoding, and previewing of Usenet newsgroup files.
//	Copyright (C) 1999,2000		David J.V. Brown, MD
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "Exceptions.h"
#include "ErrorLogger.h"
#include "HtmlGenerator.h"
#include <string>
#include <iostream>
#include <stdio.h>

using std::cerr;
using std::string;
using namespace exc;

///////////////////////////////////////////////////////////////////////////////////

Basic::Basic(	const string& location = "Unknown",
				const string& condition = "Unknown"	)
	: abort(false), loc( location ), cond( condition )	
{
}

Basic::~Basic() 
{}

void Basic::handle()
{	
	errmsg +="EXCEPTION:\n\tError: ";
	errmsg += cond + "\n\tMethod: " + loc + "\n";
	
	exc_specific();		// executes code specific to the exception subclass thrown
						// (example of Template Method design pattern)

	ErrorLogger::Instance()->client_said( errmsg );
	ErrorLogger::Instance()->dump_to_file();
	cerr << errmsg;
	cerr << endl;
	
	if (abort)
		exit(1);
}

void Basic::exc_specific() 
// MAY be overridden in subclasses to provide additional information about the error
// or take specific action to remedy the problem.
{ 
	errmsg += "\n\tNo additional information available.";
}


////////////////////////////////////////////////////////////////////////////////////

Fatal::Fatal( 	const string& location = "Unknown",
				const string& condition = "Fatal Error"	)
	: Basic( location, condition )
{
}

Fatal::~Fatal() {}

void Fatal::exc_specific()
{
	errmsg += "\nAborting...\n";
	abort = true;
}

////////////////////////////////////////////////////////////////////////////////////

NullPointer::NullPointer( const string& location = "Unknown" )
	: Fatal( location, "Attempted to dereference or access a null pointer." )
{}

NullPointer::~NullPointer() {}

void NullPointer::exc_specific()
{
	errmsg += "\nUnable to continue; aborting.\n";
	abort = true;
}

////////////////////////////////////////////////////////////////////////////////////
FileAccess::FileAccess( const string& location = "Unknown",
						const string& filename = "Filename Not Specified" )
	: Basic( location, "File Not Found or Access Denied" ), fname( filename )
{}

FileAccess::~FileAccess() {}

void FileAccess::exc_specific()
{
	errmsg += "\tFilename: ";
	errmsg += fname + '\n';
}

///////////////////////////////////////////////////////////////////////////////////
ConnectFail::ConnectFail( const string& location, const char *dotquad, int port_num )
	: Basic( location, "Could not connect to remote host" ), dquad( dotquad ),
	port( port_num )
{}

void ConnectFail::exc_specific()
{
	char port_str[8];
	snprintf( port_str, 8, "%d", port );
	
	errmsg += "\tIP Address: ";
	errmsg += ((dquad + "\n\tPort: ") + port_str) + "\n";
	cerr << "The server may be down, or you may be experiencing network problems."
		 << endl;
}

///////////////////////////////////////////////////////////////////////////////////
ResolveFail::ResolveFail( const string& location, const string& address )
	: Basic( location, "Could not resolve hostname" ), addr( address )
{}

void ResolveFail::exc_specific()
{
	errmsg += "\tUnresolved Address: ";
	errmsg += addr + "\n";
	cerr 	<< "Make sure you are connected to the network and your DNS is available."
			<< endl;
}

////////////////////////////////////////////////////////////////////////////////////
ReadFail::ReadFail( const string& location )
	: Basic( location, "Failed to read from socket" )
{}
////////////////////////////////////////////////////////////////////////////////////
WriteFail::WriteFail( const string& location )
	: Basic( location, "Unable to write to socket" )
{}
////////////////////////////////////////////////////////////////////////////////////

NNTP_Error::NNTP_Error( const string& location, const string& errorline )
	: Basic(location,"NNTP Protocol Error reported by server"), err( errorline )
{}

void NNTP_Error::exc_specific()
{
	errmsg += "\tError reported: ";
	errmsg += err + "\n";
}

////////////////////////////////////////////////////////////////////////////////////

TimeOut::TimeOut( 	const string& location, HtmlGenerator *htmlgen,
					const string& pending_io = "(Err During Read)"	)
		: Basic( location, "Socket I/O Timed Out" ), pend_io( pending_io ),
		p_htgen( htmlgen )
{}

void TimeOut::exc_specific()
{
	// write out the html previews before we die!
	if (p_htgen != 0)
		p_htgen->finish();
	errmsg += "\tPending I/O: ";
	errmsg += pend_io + "\n";
}
