// HtmlGenerator.cpp

//	PicMonger automates retrieval, decoding, and previewing of Usenet newsgroup files.
//	Copyright (C) 1999,2000		David J.V. Brown, MD
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "HtmlGenerator.h"
#include "Exceptions.h"
#include "PreviewPage.h"
#include "HtmlOptionState.h"
#include "imlib.h"

#include <string>
#include <list>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <unistd.h>		// need unlink()

using namespace std;

HtmlGenerator::HtmlGenerator(const string& path)
	: p_current( 0 ), currentFile(""), p_last(0), lastFile(""), htmlpath( path ), 
	logopath(""), bgcolor(""), groupname(""), basename("preview"),
	min_height( 300 ), min_width( 300 ),
	htmlcount(1)
{
	HtmlOptionState *p_htOpts = HtmlOptionState::Instance();

	// set options according to HtmlOptionState (global) object
	img_per_row = p_htOpts->get_cols();
	row_per_page = p_htOpts->get_rows();
	t_height = p_htOpts->get_t_height();
	t_width = p_htOpts->get_t_width();
	showLogo = p_htOpts->get_show_banner();
	showGroup = p_htOpts->get_show_groupname();
	showFilenames = p_htOpts->get_show_filenames();
	sortOutput = p_htOpts->get_sort_files();
}

HtmlGenerator::~HtmlGenerator()
{
	clear_pages();
	clear_files();
}

void HtmlGenerator::add(const string& datafile)
{
	// strip the trailing '\r' from the filename, if it exists
	string fixed;
	int rfound = datafile.find("\r");
	if ( rfound >= 0 )
		fixed = datafile.substr( 0, rfound );
	else
		fixed = datafile;

	// look to see if this file is already on the list
	list<string>::iterator iter = find(files.begin(), files.end(), fixed);

	if ( iter != files.end() )	// we already have this file in the list
		return;					// so don't add it again
	else
	{
		// check to see if we meet the size minimums
		Gdk_Imlib::Image img( fixed );

		if ( (img.rgb_height() > min_height) && (img.rgb_width() > min_width) )
			files.push_back( fixed );
		else
			unlink( fixed.c_str() );
	}
}

void HtmlGenerator::finish()
{
	if (!files.empty())
	{
		// sort the output if desired
		if (sortOutput) files.sort();
		
		// create all the pages
		list<string>::iterator iter = files.begin();
		
		while (iter != files.end())
		{
			if (pages.size() == 0)	// first addition only
			{
				currentFile = basename + html_num() + ".html";
				
				p_current = new PreviewPage( htmlpath, currentFile, groupname);
				p_current->set_show_banner( showLogo );
				p_current->set_show_pagenum( true );
				p_current->set_show_group ( showGroup );
				p_current->set_bg_image( "" );
				p_current->set_bg_color( bgcolor );
				p_current->set_pagenum( html_num() );
				p_current->set_banner_path( logopath );	
				p_current->set_rows( row_per_page );
				p_current->set_cols( img_per_row );
				p_current->set_show_filenames( showFilenames );
				p_current->set_thumb_size( t_height, t_width );
				
				p_current->add_item( *iter );
				pages.push_back( p_current );
				return;
			}

			if (p_current->is_full())	// keep reference to completed page
										// for linking and generate a new page
			{
				htmlcount++;
				lastFile = currentFile;
				currentFile = basename + html_num() + ".html";
				
				p_last = p_current;
				
				p_current = new PreviewPage( htmlpath, currentFile, groupname);
				p_current->set_show_banner( showLogo );
				p_current->set_show_pagenum( true );
				p_current->set_show_group ( showGroup );
				p_current->set_bg_image( "" );
				p_current->set_bg_color( bgcolor );
				p_current->set_pagenum( html_num() );
				p_current->set_banner_path( logopath );	
				p_current->set_rows( row_per_page );
				p_current->set_cols( img_per_row );
				p_current->set_show_filenames( showFilenames );
				p_current->set_thumb_size( t_height, t_width );
						
				pages.push_back( p_current );

				// add "next" link to page just completed
				LinkType next;
				next.url = currentFile;
				next.description = "Next";
				p_last->add_link( next );

				// add "prev" link to newly created page
				LinkType prev;
				prev.url = lastFile;
				prev.description = "Previous";
				p_current->add_link( prev );
			}

			if ( iter != files.begin() )	// keeps us from adding the 1st one 2x
				p_current->add_item( *iter );
			iter++;
		}

		// now write the created pages to disk
		list<PreviewPage *>::iterator piter = pages.begin();
		while (piter != pages.end())
		{
			try {
				(*piter)->write();
			} catch (exc::FileAccess& e) {
				e.handle();
			}
			piter++;
		}
	}
	
	clear_pages();
	clear_files();
}

string HtmlGenerator::html_num()
{
	char number[6];
	char *p_number = number;
	
	snprintf( p_number, 6, "%d", htmlcount );
	string retstr = number;

	return retstr;
}

void HtmlGenerator::clear_pages()
{
	// free all the pages in the current list
	if (!pages.empty())
	{
		list<PreviewPage *>::iterator iter = pages.begin();
		while (iter != pages.end())
		{
			delete *iter;
			iter++;
		}
		pages.clear();
	}
	
	// clear references to these pages that are deleted above
	p_current = 0;
	p_last = 0;

	// reset vars to initial state
	htmlcount = 1;
	currentFile = "";
	lastFile = "";
}

void HtmlGenerator::clear_files()
{
	files.clear();
}

void HtmlGenerator::set_logopath( const string& lpath ) { logopath = lpath; }
void HtmlGenerator::set_bgcolor( const string& bg ) { bgcolor = bg; }
void HtmlGenerator::set_groupname( const string& gn ) { groupname = gn; }
void HtmlGenerator::set_basename( const string& bn ) { basename = bn; }
