//	InetSocket.cpp
//	
//	Implementation of a socket in CPP using PF_INET protocol family.
//	

//	PicMonger automates retrieval, decoding, and previewing of Usenet newsgroup files.
//	Copyright (C) 1999,2000		David J.V. Brown, MD
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "InetSocket.h"
#include "Exceptions.h"

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>

InetSocket::InetSocket(int type=SOCK_STREAM, int protocol=0)
	: type_(type), protocol_(protocol)
{
	local_fd = socket( PF_INET, type_, protocol_ );
	
	if (local_fd < 0)
		perror("InetSocket(int, int)");
}

int InetSocket::s_connect( const char *dotquad, int port ) 
throw (exc::ConnectFail)
{
	// construct remote socket address to which we try to connect 
	remote_addr.sin_addr.s_addr = inet_addr( dotquad );
	remote_addr.sin_port = htons(port);
	remote_addr.sin_family = AF_INET;
	remote_addr_len = sizeof( remote_addr );
		 
	// try to connect (this) local socket to it
	int result = connect( local_fd, (struct sockaddr *) &remote_addr, remote_addr_len);
	if ( result < 0 )
	{
		throw exc::ConnectFail( "InetSocket::s_connect()", dotquad, port );
	}
	return result;
}

int InetSocket::s_bind( int port, const char *dotquad = (const char *)0)
{
	// construct local address to which we want our unnamed socket bound
	if (!dotquad)
		local_addr.sin_addr.s_addr = htonl( INADDR_ANY );
	else
		local_addr.sin_addr.s_addr = inet_addr( dotquad );

	local_addr.sin_port = htons(port);
	local_addr.sin_family = AF_INET;
	local_addr_len = sizeof( local_addr );

	int result = bind( local_fd, (struct sockaddr *) &local_addr, local_addr_len);
	if (result < 0)
		perror("s_bind()");

	return result;
}

int InetSocket::s_listen( int q_len )
{
	int result = listen( local_fd, q_len );
	if ( result < 0)
		perror("InetSocket::s_listen()");
	return result;
}

int InetSocket::s_accept()
{
	remote_fd = accept( local_fd, (struct sockaddr *)&remote_addr, &remote_addr_len);
	if (remote_fd < 0)
		perror("InetSocket::s_accept()");
	return remote_fd;
}

int InetSocket::s_read( int read_fd, void *data, int read_max ) 
throw (exc::ReadFail)
{
	int nread = read(read_fd, data, read_max);
	
	if ( nread < 0)
		throw exc::ReadFail( "InetSocket::s_read()" );
		
	return nread;
}

int InetSocket::s_write( const void *data, int write_max )
throw (exc::WriteFail)
{
	int result = write(local_fd, data, write_max);
	
	if ( result < 0 )
		throw exc::WriteFail( "InetSocket::s_write()" );
	return result;
}

int InetSocket::s_close()
{
	return close( local_fd );
}
