// NewsScanner.cpp

//	PicMonger automates retrieval, decoding, and previewing of Usenet newsgroup files.
//	Copyright (C) 1999,2000		David J.V. Brown, MD
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//define DEBUG
//define TRACE

#include "FilenameCleaner.h"
#include "NewsScanner.h"
#include "NewsScannerView.h"
#include "InetSocket.h"
#include "HtmlGenerator.h"
#include "ErrorLogger.h"
#include "GroupState.h"
#include "HtmlOptionState.h"
#include "AuthTab.h"

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>

#include <stdlib.h>

#include <sys/time.h>		// needed for select()
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string>
#include <iostream>
#include <fstream>

#define TIMEOUT_SECS 60

#ifndef PKGDATADIR
	#define PKGDATADIR "/usr/local/share/PicMonger"
#endif

using namespace std;


NewsScanner::NewsScanner(NewsScannerView *nsv, const char* saddr, 
	const char* ng, ScanOrder so=NS_FORWARD)

	: 	p_NSV(nsv), so_(so), htmlgen("."), p_htgen( 0 ),
		serv_addr( saddr ), serv_port( 119 ), group( ng ),
	  	numToScan(0), numScanned(0), bytesSoFar(0), status_(""), msgHasData(false),
	  	numDataFiles(0), suicide( false )
{
	sock = new InetSocket;
	p_cleaner = new FilenameCleaner();

	if (p_NSV->doPreview())
	{
		// (currently) unmodifiable "options"
		htmlgen.set_logopath("file:" PKGDATADIR "/pmlogo_big.jpg");
		htmlgen.set_groupname( group );

		p_htgen = &htmlgen;			// p_htgen is used when certain exceptions
									// are thrown so previews are written before
									// the scanner dies.  It's initialized to NULL
									// above, and only set to a usable value if
									// HTML previews are turned on.
	}
	
	statdata.resp_code = "";
	statdata.numposts = "";
	statdata.firstmsg = "";
	statdata.lastmsg = "";
}

NewsScanner::~NewsScanner()
{
	finish();
	sock->s_close();
	delete sock;
	delete p_cleaner;
}

void NewsScanner::shutdown() 
{
	suicide = true;
}

void NewsScanner::check_suicide() throw (exc::ScannerShutdown)
{
	if (suicide)
	{
		suicide = false;
		throw exc::ScannerShutdown();
	}
}

int NewsScanner::establish() throw (exc::ResolveFail)
// establish a connection with the server and get hello text
{
	status_ = "Connecting to server...";
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}
	
	hostent *hostinfo;
	char **addrs;
	char *dotquad;
	
	hostinfo = gethostbyname( serv_addr.c_str() );
	if (!hostinfo) 
	{
		throw exc::ResolveFail( "NewsScanner::establish()",
								serv_addr );
	}

	addrs = hostinfo->h_addr_list;
	dotquad = inet_ntoa( *(struct in_addr *) *addrs);  // get dotted quad from URL

	int result = sock->s_connect( dotquad, serv_port);
	
	get_response(false);
	
	bool NNTPerr = is_NNTP_error();

	bool needAuth = AuthTab::Instance()->get_use_auth();
	
	if (needAuth)
	{
		string userCmd = "AUTHINFO USER ";
		string passCmd = "AUTHINFO PASS ";
		userCmd += AuthTab::Instance()->get_user();
		passCmd += AuthTab::Instance()->get_pass();

		// send user authentication info
		send_command( userCmd.c_str(), false);
		get_response(false);
		send_command( passCmd.c_str(), false);
		get_response(false);
	}
	
	result = (NNTPerr) ? 1 : 0;	// return 1 if error, 0 if OK
	return result;
}

int NewsScanner::choose_group()
// send GROUP command to select newsgroup, get response and parse info on current posts
{
	status_ = "Getting post data for newsgroup...";
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}	
	
	string group_cmd("GROUP ");
	group_cmd += group;

	send_command(group_cmd.c_str(), false);
	
	get_response(false);
	bool NNTPerr = is_NNTP_error();
	
	if (NNTPerr)
	{
		return 1;					
	} 													
	else 							
	{
		parse_status();
		status_ = "Newsgroup statistics parsed.";
		try 
		{
			Notify();
		} 
		catch(exc::NullPointer& e) 
		{
			e.handle();
		}

		return 0;	
	}
			
}

void NewsScanner::send_command(const char* cmdline, bool ttyout)
throw (exc::TimeOut)
{
	struct timeval timeOut;
	timeOut.tv_sec = TIMEOUT_SECS;				// 60 seconds until server timeout
	timeOut.tv_usec = 0;
	
	int sockfd = sock->get_fd();
	static const char* EOC = "\r\n";				// end of command
	
	FD_ZERO( &fd_write );
	FD_SET( sockfd, &fd_write);
	FD_ZERO( &fd_err );
	FD_SET( sockfd, &fd_err );

	status_="Waiting for server to accept data...";
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}
	
	int fd_count = select( sockfd+1, 0, &fd_write, &fd_err, &timeOut ); 
	if (fd_count == 0)
		throw exc::TimeOut("NewsScanner::send_command", p_htgen, cmdline);
		
	if ( FD_ISSET(sockfd, &fd_write) )
	{
		sock->s_write( cmdline, strlen(cmdline) );
		sock->s_write( EOC, strlen( EOC ) );
		
		ErrorLogger::Instance()->client_said( cmdline );
		
		if (ttyout) cout << "Command sent to server: " << cmdline << endl;
	}
	else 
	{
		if ( FD_ISSET(sockfd, &fd_err) ) 		// socket error detected
		{
			cerr << "send_command(): socket error detected" << endl;
			sock->s_close();
			exit(1);
		}
	}
}

int NewsScanner::get_response(bool ttyout)
throw (exc::NNTP_Error, exc::TimeOut)
{
	struct timeval timeOut;
	timeOut.tv_sec = TIMEOUT_SECS;	// give server 60 seconds to respond
	timeOut.tv_usec = 0;
	
	int sockfd = sock->get_fd();
	int nread = -1;

	FD_ZERO( &fd_read ); FD_SET( sockfd, &fd_read );
	FD_ZERO( &fd_err ); FD_SET( sockfd, &fd_err );
	
	int fd_count = select( sockfd+1, &fd_read, 0, &fd_err, &timeOut);	
	if ( fd_count == 0 ) 	// timed out
		throw exc::TimeOut("NewsScanner::get_response()", p_htgen);
		
	if ( FD_ISSET( sockfd, &fd_read ) ) 		// socket ready to be read
	{
		if ( (nread = sock->s_read( sockfd, &response, RESP_SIZE )) < 0 )
		{
			perror("InetSocket::s_read()");
			status_ = "Error trying to read socket!";
			try {
				Notify();
			} catch(exc::NullPointer& e) {
				e.handle();
			}
			sock->s_close();
			delete this;	// warn observers of death and vanish!
		}
	}
	else
	{
		if ( FD_ISSET( sockfd, &fd_err ) )
		{
			perror("socket error detected");
			sock->s_close();
			exit(1);
		}
	}

	response[nread] = 0;		// add trailing null to everything
	
	ErrorLogger::Instance()->server_said( response );
	

	if (is_NNTP_error())
		throw exc::NNTP_Error( "NewsScanner::get_response()", response );
		
	if (ttyout) cout << response << endl;

	bytesSoFar += nread;
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}
	
	return nread;			// return number of bytes read succesfully
}

bool NewsScanner::process_message(bool ttyout)
throw (exc::FileAccess, exc::ScannerShutdown)
{
	string inbuffer;
	int find_result = 0;
	
	static const char *endFlag = ".\r\n";
	static const char *uueBegin = "\nbegin ";
	static const char *uueEnd = "\nend";
	static const char *mimeHeader = "Content-Transfer-Encoding:";
	static const char *base64data = "/9";

	bool endFound = false;
	bool isUUE = false;
	bool isMIME = false;
	bool hasUUEBegin = false;
	bool hasMIMEhead = false;
	bool isBase64 = false;

	fstream fout("tempdata.txt", ios::out);
	if (!fout)
		throw exc::FileAccess(	"NewsScanner::process_message()",
								"tempdata.txt" );
	
	string fixed_fn;		// holds the repaired/validate filename
	string old_fn;

	msgHasData = false;
	bytesSoFar = 0;
	
	status_ = "Scanning for data markers...";
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}
	
	do
	{
#ifdef TRACE
		cerr << "Getting response..." << endl;
#endif

		try { 
			get_response(false);
		 	inbuffer = response;
		}
		catch( exc::Basic& except ) {
			except.handle();
			throw exc::ScannerShutdown();
			return false;
		}
		
#ifdef TRACE
		cerr << "Setting inbuffer..." << endl;
#endif
		p_cleaner->set_inbuffer( inbuffer );
		
		hasMIMEhead = ( (find_result = inbuffer.find(mimeHeader)) >= 0);
		isBase64 = ( (find_result = inbuffer.find(base64data)) >= 0);
		hasUUEBegin = ((find_result = inbuffer.find(uueBegin)) >= 0);
		
		if ( hasMIMEhead && isBase64 && (!msgHasData) )
		{
#ifdef TRACE
			cerr << "Starting MIME filename processing..." << endl;
#endif
			if (ttyout) cout << "Getting MIME encoded attachment";
#ifdef TRACE
			cerr << "Getting MIME filename..." << endl;
#endif
			old_fn = p_cleaner->get_MIME_filename();
			fixed_fn = old_fn;
#ifdef TRACE
			cerr << "Validating filename..." << endl;
#endif
			msgHasData = p_cleaner->validate_filename( fixed_fn );		
			// mangles fixed_fn to be ext2fs-friendly as a side effect
			
			isMIME = msgHasData;
			if (isMIME)
			{
				numDataFiles++;
				status_ = "Getting MIME file: ";
				status_ += fixed_fn;
				try {
					Notify();
				} catch(exc::NullPointer& e) {
					e.handle();
				}
#ifdef TRACE
				cerr << "Fixing filaname in MIME buffer..." << endl;
#endif
				inbuffer = p_cleaner->fix_MIME_buffer( old_fn, fixed_fn );
			}
			else
			{
				status_ = "Unwanted filetype. Scanning through...";
				try {
					Notify();
				} catch(exc::NullPointer& e) {
					e.handle();
				}
			}
		}
		else
		if ( (hasUUEBegin) && (!msgHasData) )
		{	// if begin flag found and first one in this message
#ifdef TRACE
			cerr << "Starting UUE processing..." << endl;
#endif
			if (ttyout) cout << "Getting UUEncoded attachment";
#ifdef TRACE
			cerr << "Getting UUE filename..." << endl;
#endif
			old_fn = p_cleaner->get_UUE_filename();
			fixed_fn = old_fn;
#ifdef TRACE
			cerr << "Validating UUE filename..." << endl;
#endif
			msgHasData = p_cleaner->validate_filename( fixed_fn ); 	
			// modifies fixed_fn!!

			isUUE = msgHasData;
			if ( isUUE )
			{
#ifdef TRACE
				cerr << "UUE filename validated..." << endl;
#endif
				numDataFiles++;
				status_ = "Getting UUEncoded file: ";
				status_ += fixed_fn;
				try {
					Notify();
				} catch(exc::NullPointer& e) {
					e.handle();
				}
#ifdef TRACE
				cerr << "Replacing UUE buffered filename..." << endl;
#endif
				// replace the encoded filename before written to local file
				inbuffer = p_cleaner->fix_UUE_buffer( old_fn, fixed_fn );
			}
			else
			{
				status_ = "Unwanted filetype. Scanning through...";
				try {
					Notify();
				} catch(exc::NullPointer& e) {
					e.handle();
				}
			}
		}

		if (msgHasData)
		{
#ifdef TRACE
			cerr << "Writing data buffer..." << endl;
#endif
			fout << inbuffer << flush;
			if (ttyout) cout << "." << flush;
			try {
				Notify();
			} catch(exc::NullPointer& e) {
				e.handle();
			}
		}

		find_result = 1;
		int scanidx = 0;
		while ((!endFound) && (find_result >=0))
		{
			// scan through the buffer for "." on a new line

			find_result = inbuffer.find( endFlag, scanidx );
			if (find_result == -1) continue;
			
			if (find_result == 0) // is the . at the beginning of the buffer?
				endFound = true; 
			else if (find_result > 0)
			{
				if (inbuffer[find_result - 1] == '\n') // is this a newline?
					endFound = true;
				else
				{
					endFound = false;
					scanidx = find_result + 1;	// keep looking further in the buffer
				}
			}
		}
			
				


		//endFound = ( (find_result = inbuffer.find( endFlag )) >= 0 );
		//if ( find_result > 0 )
		// ".\r\n" was found in the middle of the packet -- make sure it's on
		// a new line and not just the end of a sentence or something.
		//{
		//	if (inbuffer[find_result-1] != '\n')
		//		endFound = false;
//#ifdef TRACE
//			cerr << "At position: " << find_result << "a suspected . end was found."
//				<< endl
//				<< "endFound was subsequently adjusted to: " << endFound << endl;
//#endif
//		}

#ifdef TRACE
		cerr << "endFlag found: " << endFound << "\t";
#endif
		if (!endFound)
		{
			endFound = ( (find_result = inbuffer.find( uueEnd )) >= 0 );
#ifdef TRACE
			cerr << "uueEnd found: " << endFound << endl;
#endif
		}
	} while ( !endFound );	
	
	numScanned++;
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}
	
	if (ttyout) cout << ( (msgHasData) ? "done." : "No encoded attachment found." ) 
		<< endl;

#ifdef TRACE
	cerr << "Closing local file..." << endl;
#endif
	fout.close();
	
	if (isUUE)
	{
		status_ = "UUDecoding message...";
		try {
			Notify();
		} catch(exc::NullPointer& e) {
			e.handle();
		}
#ifdef TRACE
		cerr << "System call to UUdecode..." << endl;
#endif
		system("uudecode < tempdata.txt 2> /dev/null");

		status_ = "Message decoded.";
		try {
			Notify();
		} catch(exc::NullPointer& e) {
			e.handle();
		}
#ifdef TRACE
		cerr << "Adding to htmlgen if desired..." << endl;
#endif
		if ( p_NSV->doPreview() ) 
			htmlgen.add( fixed_fn );
	}
	else 
	if (isMIME)
	{
		status_ = "Decode MIME attachment...";
		try {
			Notify();
		} catch(exc::NullPointer& e) {
			e.handle();
		}
#ifdef TRACE
		cerr << "System call to munpack..." << endl;
#endif
		system("munpack -q tempdata.txt 2> /dev/null");

		status_ = "MIME attachment decoded.";
		try {
			Notify();
		} catch(exc::NullPointer& e) {
			e.handle();
		}
#ifdef TRACE
		cerr << "Adding HTML to preview if tired..." << endl;
#endif
		if ( p_NSV->doPreview() )
			htmlgen.add( fixed_fn );
	}
#ifdef TRACE
	cerr << "Returning from process_message() with: " << msgHasData << endl;
#endif
	return (msgHasData);	// returns true if Encoded data was processed
}

void NewsScanner::parse_status()
// parse the response after a group is selected
{
	char *running = 0;
	char *token = 0;

#ifdef TRACE
	cout << "parse_status: response = " << response << endl;
#endif

	running = response;
	for (int i=0; i<4; i++)
	{
		token = strsep( &running, " " );
		switch (i)
		{
			case 0:
				statdata.resp_code = token;
				break;
			case 1:
				statdata.numposts = token;
				break;
			case 2:
				statdata.firstmsg = token;
				break;
			case 3:
				statdata.lastmsg = token;
				break;
			default:
				cerr << "out of switch bounds in NewsScanner::parse_status()\n";
		}
	}
}

bool NewsScanner::scan() 
// returns true on successful, complete scan.  Otherwise false.
{
	int protocolErr = 0;
	
	if ( (protocolErr = establish()) == 1) return false;
	if ( (protocolErr = choose_group()) == 1) return false;

	GroupState gState(group, statdata.lastmsg);
	string lastread = gState.get_last_message();
	
	string statcmd("STAT ");
	if ( lastread == "" )
		statcmd += statdata.firstmsg;
	else
		statcmd += lastread;

	if ( so_ == NS_REVERSE )
		nextcmd = "LAST";
	else
		nextcmd = "NEXT";
	
	send_command( statcmd.c_str(), false );
	get_response(false);

	next_message();
	gState.set_last_message( current_msg_id );
	
	sscanf( statdata.numposts.c_str(), "%d", &numToScan );
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}

	bool abortingScan = false;
	try 
	{
		while (!eog()) // not end of group
		{
			check_suicide();
			send_command("BODY", false);
			
			try {
				process_message(false);
			} catch( exc::FileAccess &e ) 	{
				e.handle();
			}
			
			if (!abortingScan) 
			{
				next_message();
				gState.set_last_message( current_msg_id );
				
				if (is_NNTP_error()) 
				{
					finish();
					return false;
				}
			}
		}
	}
	catch (exc::ScannerShutdown)
	{
		abortingScan = true;
	}

	status_ = (abortingScan) ? "Aborting current scan job...\n" : "";
	status_ += "QUIT command sent to server.";
	
	try {
		Notify();
	} catch(exc::NullPointer& e) {
		e.handle();
	}

	send_command("QUIT", false);
	finish();
	return (!abortingScan);
}

void NewsScanner::next_message()
{
	send_command( nextcmd.c_str(), false );
	get_response( false );
	
	if ( !eog() )
	{
		string temp( response );
		int first_space = temp.find(" ");
		int second_space = temp.find(" ", first_space+1);
		current_msg_id = temp.substr( first_space+1, second_space-(first_space+1) );
	}
}
	

void NewsScanner::finish()
{
	if (p_NSV->doPreview())
		htmlgen.finish();
	try {
		Detach(p_NSV);
	} catch (exc::NullPointer& e) {
		e.handle();
	}
}

bool NewsScanner::is_NNTP_error()
{
	static const int nErrors = 6;
	static const char *errCodes[nErrors] = 
	{ 	"400 ",	// service discontinued by server 
		"500 ", // command not recognized
		"501 ", // command syntax error
		"502 ", // access restricted / permission denied
		"503 ",	// program fault -- command not performed
		"411 "	// no such newsgroup available	
	};

	for (int i=0; i<nErrors; i++)
	{
		if ( strncmp( response, errCodes[i], 4) == 0 )	// error matched
		{
			status_ = "NNTP error: ";
			status_ += response;
			try {
				Notify();
			} catch(exc::NullPointer& e) {
				e.handle();
			}
			return true;
		}
	}

	return false;	// no matching error code found
}

bool NewsScanner::eog()
{
	// we won't treat these as errors... rather "end of group" flags...
	// 421 and 422 are the "no next article" and "no previous article" 
	// NNTP error codes...
	
	return	(	(strncmp(response, "422 ", 4) == 0) ||
				(strncmp(response, "421 ", 4) == 0)		);
}

float NewsScanner::pct_scanned() const
{
	// Calculate an approximate percent of newsgroup already scanned, 
	// including any messages scanned in prior sessions.
	
	// returns 0.0 for any computed values x such that x < 0.0 or x > 1.0
	// Gtk::ProgressBar complains if these constraints are not met.
	
	float first, last, current;

	// convert string message # data to float
	sscanf( statdata.firstmsg.c_str(), "%f", &first );
	sscanf( statdata.lastmsg.c_str(), "%f", &last );
	sscanf( current_msg_id.c_str(), "%f", &current );

	// returns 0.0 for any computed values x such that x < 0.0 or x > 1.0
	// Gtk::ProgressBar complains if these constraints are not met...
	
	if ( (last - first) == 0.0 )		// avoid divide by zero
		return 0.0;
	else
	{
		float result = (current-first)/(last-first);
		if ( (result < 0.0) || (result > 1.0) )
			return 0.0;
		else
			return result;
	}	
}