// NewsScanner.h
//
// A NewsScanner is an object that scans a specified 
// Usenet newsgroup for UUEncoded binary files, retrieves
// them, and decodes them in the working directory.
//
// A NewsScanner is also derived from class Subject, meaning that
// a class derived from class Observer can register with it and be updated
// automagically.
// 

//	PicMonger automates retrieval, decoding, and previewing of Usenet newsgroup files.
//	Copyright (C) 1999,2000		David J.V. Brown, MD
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#ifndef _NEWS_SCANNER_H
#define _NEWS_SCANNER_H

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string>
#include "Subject.h"
#include "HtmlGenerator.h"
#include "Exceptions.h"

#define RESP_SIZE 8192

using std::string;

enum ScanOrder { NS_FORWARD, NS_REVERSE };


class InetSocket;
class NewsScannerView;
class FilenameCleaner;

class NewsScanner : public Subject {
	public:
		NewsScanner(NewsScannerView *nsv, const char* url, 
				const char* newsgroup, ScanOrder so=NS_FORWARD);
		~NewsScanner();

		// GET type functions for observers to use when updating
		inline string server_addr() const
			{ return serv_addr; }
		inline string group_name() const
			{ return group; }
		inline string status() const
			{ return status_; }
		inline int num_to_scan() const
			{ return numToScan; }
		float pct_scanned() const;
		inline unsigned long bytes_so_far()
			{ return bytesSoFar; }
		inline bool msg_has_data() const
			{ return msgHasData; }
		inline int num_data_files() const
			{ return numDataFiles; }
		
		bool scan();
		void finish();
		void shutdown();
		
	private:
		NewsScannerView *p_NSV;
		FilenameCleaner *p_cleaner;
		
		ScanOrder so_;

		HtmlGenerator htmlgen;
		HtmlGenerator *p_htgen;
		InetSocket *sock;
		fd_set fd_read;
		fd_set fd_write;
		fd_set fd_err;
		
		char response[RESP_SIZE]; 

		int establish() throw (exc::ResolveFail);
		int choose_group();
		
		void next_message();
		bool is_NNTP_error();
		bool process_message(bool ttyout) throw (exc::FileAccess, exc::ScannerShutdown);
		
		void send_command(const char* cmd, bool ttyout) throw (exc::TimeOut);
		int get_response(bool ttyout) throw (exc::NNTP_Error, exc::TimeOut);
		
		struct StatData {
			string resp_code;
			string numposts;
			string firstmsg;
			string lastmsg;
		} statdata;
		void parse_status();

		string serv_addr;
		int serv_port;
		
		string group;
		
		int numToScan;
		int numScanned;
		unsigned long bytesSoFar;
		string status_;
		bool msgHasData;
		int  numDataFiles;
		
		string nextcmd;
		string current_msg_id;

		bool eog();
		bool suicide;
		void check_suicide() throw (exc::ScannerShutdown);
		
};
#endif
